==> mx2g_1.1.1.tar.gz: mx-1.1.1/fma/libfma/lf_clos_level.c <==
/*
 * Routing routines
 */

#include <stdio.h>

#include "libfma.h"
#include "lf_internal.h"
#include "lf_fabric.h"
#include "lf_clos_level.h"


#define FOR_EACH_XBAR(FP,XP)				\
  {							\
    int __x;						\
    for (__x=0; __x<FP->num_xbars; ++__x) {		\
      XP = FP->xbars[__x];				\
      {
#define END_XBAR_LOOP }}}

#define FOR_EACH_NIC(FP, HP, NP)			\
  {							\
    int __h;						\
    for (__h=0; __h<FP->num_hosts; ++__h) {		\
      int __n;						\
      HP = FP->hosts[__h];				\
      for (__n=0; __n<HP->num_nics; ++__n) {		\
	NP = HP->nics[__n];				\
	{
#define END_NIC_LOOP }}}}

/*
 * local prototypes
 */
static void lf_find_min_nic_dist(struct lf_xbar *xp);
static void lf_find_max_nic_dist(struct lf_xbar *xp);
static int lf_find_clos_depth(struct lf_fabric *fp);
static int lf_propagate_clos_levels(struct lf_fabric *fp);
static int lf_clos_alloc_space(struct lf_fabric *fp);
static void lf_clos_free_space(struct lf_fabric *fp);

/*
 * Label all the xbars with their "clos level".  The spine is defined to be
 * the set of all xbars sharing the minimum maximum distance to any node.
 * That is, for each xbar, find the maximum distance to any node, and call it
 * "D".  The set of xbars sharing this minimum value D is the spine. Label
 * them all as level D, then every directly connected unlabelled xbar as
 * D-1, etc.  In a clos, no xbar of level L should be connected to any xbars
 * with level other than L, L+1, or L-1.
 */
int
lf_clos_label_xbars(
  struct lf_fabric *fp)
{
  struct lf_xbar *xp;
  int alloced;
  int rc;

  rc = 0;
  alloced = FALSE;

  /* quick trip if no xbars */
  if (fp->num_xbars == 0) return 0;

  /* Attach some scratch memory to each xbar for us to use */
  rc = lf_clos_alloc_space(fp);
  if (rc == -1) LF_ERROR(("Error allocating scratch xbar space"));
  alloced = TRUE;

  /* zero out everyone's max distance */
  FOR_EACH_XBAR(fp, xp)
    xp->min_nic_dist = -1;
    LC_XBAR(xp)->max_nic_dist = 0;
    xp->clos_level = 0;
    LC_XBAR(xp)->last_visitor = NULL;
  END_XBAR_LOOP

  /* find closest and farthest NIC to each xbar */
  FOR_EACH_XBAR(fp, xp)
    lf_find_min_nic_dist(xp);
  END_XBAR_LOOP

  /* clear last visitors, and now find the max distance to any NIC */
  FOR_EACH_XBAR(fp, xp)
    LC_XBAR(xp)->last_visitor = NULL;
  END_XBAR_LOOP
  FOR_EACH_XBAR(fp, xp)
    lf_find_max_nic_dist(xp);
  END_XBAR_LOOP

  /* Find the clos depth */
  LC_FABRIC(fp)->clos_depth = lf_find_clos_depth(fp);

  /* Propagate clos_depth to all xbars */
  rc = lf_propagate_clos_levels(fp);
  if (rc == -1) LF_ERROR(("Error propagsting clos levels"));

  {
    int *links;

    LF_CALLOC(links, int, LC_FABRIC(fp)->clos_depth+1);

    FOR_EACH_XBAR(fp, xp)
      int l;
      int p;

      l = xp->clos_level;

      /* validate l, as this is just informational */
      if (l > LC_FABRIC(fp)->clos_depth || l < 0) {
	LF_ERROR(("Bad clos-level on xbar, not a clos?"));
	continue;
      }

      /* check every xbar neighbor */
      for (p=0; p<xp->num_ports; ++p) {
	struct lf_xbar *xp2;

	xp2 = LF_XBAR(xp->topo_ports[p]);
	if (xp2 != NULL
	    && (xp2->n.type == LF_NODE_NIC
	        || (xp2->n.type == LF_NODE_XBAR
		    && xp2->clos_level < l))) {
	  ++links[l];
	}
      }
    END_XBAR_LOOP

    LF_FREE(links);
  }

 except:
  if (alloced) lf_clos_free_space(fp);
  return rc;
}

/*
 * Find the distance to the closest NIC from this xbar
 */
static void
lf_find_min_nic_dist(
  struct lf_xbar *xbar)
{
  int dist;
  struct lf_xbar *xp;
  struct lf_xbar *work_list;

  dist = 1;

  /* initialize work_list to just be first xbar */
  work_list = xbar;
  LC_XBAR(xbar)->bf_next = NULL;
  LC_XBAR(xbar)->last_visitor = xbar;

  /* Breadth-first until nothing more to do */
  while (work_list != NULL) {

    /* save this list and start building new work_list */
    xp = work_list;
    work_list = NULL;

    /* process every xbar in old work_list */
    while (xp != NULL) {
      int p;

      /* Check all ports for NICs and unvisited xbars */
      for (p=0; p<xp->num_ports; ++p) {
	struct lf_node *np2;

	np2 = xp->topo_ports[p];
	if (np2 == NULL) continue;

	if (np2->type == LF_NODE_NIC) {
	  xbar->min_nic_dist = dist;
	  return;

	} else if (np2->type == LF_NODE_XBAR
		   && LC_XBAR(np2)->last_visitor != xbar) {

	  /* add to work_list */
	  LC_XBAR(np2)->bf_next = work_list;
	  work_list = LF_XBAR(np2);
	  
	  /* mark as visited to prevent multiple visitors */
	  LC_XBAR(np2)->last_visitor = xbar;
	}
      }

      /* next in list */
      xp = LC_XBAR(xp)->bf_next;
    }

    /* increment distance for each go-round */
    ++dist;
  }
}

/*
 * Find the longest distance from this xbar to any NIC, going only "up",
 * "down", or "up then down" - do not go "down then up".  Down is defined as
 * going from an xbar to another xbar with min_nic_dist <= the current xbar.
 * Up is going to an xbar with a min_nic_dist greater than the current xbar.
 * The memory in the algorithm is accomplished by having an "up_list" and a
 * "down_list".  Both lists are processed each time through the breadth-
 * first loop, and we allow both up and down connections from the up list,
 * but only down connections from the down list.
 */
static void
lf_find_max_nic_dist(
  struct lf_xbar *xbar)
{
  int dist;
  struct lf_xbar *xp;
  enum { UP, DOWN, FLAT, NUM_DIRS };
  struct lf_xbar *work_list[NUM_DIRS];
  struct lf_xbar *new_work_list[NUM_DIRS];

  dist = 1;

  /* initialize down work_list to just be first xbar */
  new_work_list[UP] = xbar;
  new_work_list[DOWN] = NULL;
  new_work_list[FLAT] = NULL;
  LC_XBAR(xbar)->bf_next = NULL;
  LC_XBAR(xbar)->last_visitor = xbar;

  /* Breadth-first until nothing more to do */
  while (new_work_list[UP] != NULL
         || new_work_list[DOWN] != NULL
         || new_work_list[FLAT] != NULL) {
    int dir;

    /* Copy over work lists */
    for (dir=0; dir<NUM_DIRS; ++dir) {
      work_list[dir] = new_work_list[dir];
      new_work_list[dir] = NULL;
    }

    /* process both lists each time through */
    for (dir=0; dir<NUM_DIRS; ++dir) {

      /* save this list and start building new work_list */
      xp = work_list[dir];
      work_list[dir] = NULL;

      /* process every xbar in old work_list */
      while (xp != NULL) {
	int p;

	/* Check all ports for NICs and unvisited xbars */
	for (p=0; p<xp->num_ports; ++p) {
	  struct lf_node *np2;

	  np2 = xp->topo_ports[p];
	  if (np2 == NULL) continue;

	  if (np2->type == LF_NODE_NIC) {
	    if (dist > LC_XBAR(xbar)->max_nic_dist) {
	      LC_XBAR(xbar)->max_nic_dist = dist;
	    }

	  } else if (np2->type == LF_NODE_XBAR
		     && LC_XBAR(np2)->last_visitor != xbar) {
	    struct lf_xbar *xp2;
	    int new_dir;

	    xp2 = LF_XBAR(np2);
	    if (xp2->min_nic_dist > xp->min_nic_dist) {
	      new_dir = UP;
	    } else if (xp2->min_nic_dist
		       < xp->min_nic_dist) {
	      new_dir = DOWN;
	    } else {
	      new_dir = FLAT;
	    }

	    /* don't go from flat or down to anything but down! */
	    if (dir != UP && new_dir != DOWN) {
	      continue;
	    }

	    /* add to new work_list */
	    LC_XBAR(xp2)->bf_next = new_work_list[new_dir];
	    new_work_list[new_dir] = xp2;
	    
	    /* mark as visited to prevent multiple visitors */
	    LC_XBAR(xp2)->last_visitor = xbar;
	  }
	}

	/* next in list */
	xp = LC_XBAR(xp)->bf_next;
      }
    }

    /* increment distance for each go-round */
    ++dist;
  }
}

/*
 * propagate clos levels.  start with all xbars whose "max_nic_dist" is
 * equal to the fabric clos depth.  Set their clos_level to be clos_depth.
 * Then, propagate out to all other xbars.  
 * We may go down then back up to tag "false spines", so we need to be
 * willing to go both "up" and "down" - if "max_distance" is greater than
 * ours, subtract one from our clos level, if greater, add one.
 * If equal, print an error because I am not sure that's possible...
 */
static int
lf_propagate_clos_levels(
  struct lf_fabric *fp)
{
  struct lf_xbar *work_list;
  int x;

  /* initialize work_list to be all xbars with max_nic_dist == clos_depth */
  work_list = NULL;
  for (x=0; x<fp->num_xbars; ++x) {
    struct lf_xbar *xp;

    xp = fp->xbars[x];
    if (LC_XBAR(xp)->max_nic_dist == LC_FABRIC(fp)->clos_depth) {
      xp->clos_level = LC_FABRIC(fp)->clos_depth;

      /* tag this as visited and add to work list */
      LC_XBAR(xp)->last_visitor = fp;
      LC_XBAR(xp)->bf_next = work_list;
      work_list = xp;
    }
  }

  /* Breadth-first until nothing more to do */
  while (work_list != NULL) {
    struct lf_xbar *xp;

    /* save this list and start building new work_list */
    xp = work_list;
    work_list = NULL;

      /* process every xbar in old work_list */
    while (xp != NULL) {
      int p;

      /* Check all ports for unvisited xbars */
      for (p=0; p<xp->num_ports; ++p) {
	struct lf_xbar *xp2;

	xp2 = LF_XBAR(xp->topo_ports[p]);

	/* if non-existant, non-xbar, or visited, skip it */
	if (xp2 == NULL
	    || xp2->n.type != LF_NODE_XBAR
	    || LC_XBAR(xp2)->last_visitor == fp) {
	  continue;
	}

	/* set neighbors' clos_level based on mine */
	if (LC_XBAR(xp2)->max_nic_dist < LC_XBAR(xp)->max_nic_dist) {
	  xp2->clos_level = xp->clos_level + 1;
	} else if (LC_XBAR(xp2)->max_nic_dist
		   > LC_XBAR(xp)->max_nic_dist) {
	  xp2->clos_level = xp->clos_level - 1;
	} else {
	  LF_ERROR(("max NIC distances cannot be equal here"));
	}

	/* add to new work_list */
	LC_XBAR(xp2)->bf_next = work_list;
	work_list = xp2;
	
	/* mark as visited to prevent multiple visitors */
	LC_XBAR(xp2)->last_visitor = fp;
      }

      /* next in list */
      xp = LC_XBAR(xp)->bf_next;
    }
  }

  /* We could cleanup a little here because in a "flat-top" clos,
   * there will end up being no "level 1" nodes.  We could find the 
   * "minimum" clos level anywhere and subtract from everything
   * to make the minimum 1, but it's not necessary yet.
   */
  return 0;

 except:
  return -1;
}

/*
 * Find the clos depth by starting from some host, and traveling to xbars
 * with decreasing "max distance", choosing each time the largest max distance
 * which is less than our own until we can go no farther.  The max_nic_distance
 * of that xbar is the depth of the clos.  This allows us to avoid finding 
 * "false spines" which are sub-spines within the overall fabric.
 */
static int
lf_find_clos_depth(
  struct lf_fabric *fp)
{
  struct lf_nic *nicp;
  struct lf_xbar *xp;

  /* just take the first connected NIC in the fabric.
   * If really a clos, any will do.
   */
  {
    int h;

    nicp = NULL;
    xp = NULL;
    for (h=0; h<fp->num_hosts; ++h) {
      if (fp->hosts[h]->num_nics > 0) {
	nicp = fp->hosts[h]->nics[0];
	xp = LF_XBAR(nicp->topo_ports[0]);
	if (xp != NULL && xp->n.type == LF_NODE_XBAR) {
	  break;
	}
      }
    }
    if (xp == NULL) return 0;
  }

  while (1) {
    struct lf_xbar *nxp;
    int max_max;
    int p;

    max_max = -1;
    for (p=0; p<xp->num_ports; ++p) {
      struct lf_node *onp;

      onp = xp->topo_ports[p];
      if (onp != NULL
	  && onp->type == LF_NODE_XBAR
	  && LC_XBAR(onp)->max_nic_dist < LC_XBAR(xp)->max_nic_dist
	  && LC_XBAR(onp)->max_nic_dist > max_max) {
	max_max = LC_XBAR(onp)->max_nic_dist;
	nxp = LF_XBAR(onp);
      }
    }

    /*
     * If no higher nodes found, then this is a spine xbar!
     */
    if (max_max == -1) {
      return LC_XBAR(xp)->max_nic_dist;

    } else {
      xp = nxp;		/* otherwise, move on to the new xbar */
    }
  }
}

/*
 * Alloacate space for working variables
 */
static int
lf_clos_alloc_space(
  struct lf_fabric *fp)
{
  struct lf_clos_level_fabric_data *fdp;
  struct lf_clos_level_xbar_data *xdp;
  struct lf_xbar *xp;

  fdp = NULL;
  xdp = NULL;

  /* allocate data space (all xbars are in one alloced array for efficiency) */
  LF_CALLOC(fdp, struct lf_clos_level_fabric_data, 1);
  LF_CALLOC(xdp, struct lf_clos_level_xbar_data, fp->num_xbars);

  /* Save old fabric user data ptr and add ours in */
  fdp->old_user = fp->user;
  fp->user = fdp;

  FOR_EACH_XBAR(fp,xp)
    xdp[__x].old_user = xp->user;
    xp->user = xdp + __x;
  END_XBAR_LOOP

  return 0;

 except:
  LF_FREE(fdp);
  LF_FREE(xdp);
  return -1;
}

/*
 * Free out temporary working space
 */
static void
lf_clos_free_space(
  struct lf_fabric *fp)
{
  struct lf_clos_level_fabric_data *fdp;
  struct lf_clos_level_xbar_data *xdp;
  struct lf_xbar *xp;

  /* Restore old fabric user data ptr and free ours */
  fdp = fp->user;
  fp->user = fdp->old_user;
  LF_FREE(fdp);

  /* first xbar has the address we will free */
  xdp = fp->xbars[0]->user;
  FOR_EACH_XBAR(fp,xp)
    xp->user = xdp[__x].old_user;
  END_XBAR_LOOP
  LF_FREE(xdp);
}

==> mx2g_1.1.4.tar.gz: mx-1.1.4/fma/libfma/lf_clos_level.c <==
/*
 * Routing routines
 */

#include <stdio.h>

#include "libfma.h"
#include "lf_internal.h"
#include "lf_fabric.h"
#include "lf_clos_level.h"


#define FOR_EACH_XBAR(FP,XP)				\
  {							\
    int __x;						\
    for (__x=0; __x<FP->num_xbars; ++__x) {		\
      XP = FP->xbars[__x];				\
      {
#define END_XBAR_LOOP }}}

#define FOR_EACH_NIC(FP, HP, NP)			\
  {							\
    int __h;						\
    for (__h=0; __h<FP->num_hosts; ++__h) {		\
      int __n;						\
      HP = FP->hosts[__h];				\
      for (__n=0; __n<HP->num_nics; ++__n) {		\
	NP = HP->nics[__n];				\
	{
#define END_NIC_LOOP }}}}

/*
 * local prototypes
 */
static void lf_find_min_nic_dist(struct lf_xbar *xp);
static void lf_find_max_nic_dist(struct lf_xbar *xp);
static int lf_find_clos_depth(struct lf_fabric *fp);
static int lf_propagate_clos_levels(struct lf_fabric *fp);
static int lf_clos_alloc_space(struct lf_fabric *fp);
static void lf_clos_free_space(struct lf_fabric *fp);

/*
 * Label all the xbars with their "clos level".  The spine is defined to be
 * the set of all xbars sharing the minimum maximum distance to any node.
 * That is, for each xbar, find the maximum distance to any node, and call it
 * "D".  The set of xbars sharing this minimum value D is the spine. Label
 * them all as level D, then every directly connected unlabelled xbar as
 * D-1, etc.  In a clos, no xbar of level L should be connected to any xbars
 * with level other than L, L+1, or L-1.
 */
int
lf_clos_label_xbars(
  struct lf_fabric *fp)
{
  struct lf_xbar *xp;
  int *links;
  int alloced;
  int rc;

  rc = 0;
  alloced = FALSE;
  links = NULL;

  /* quick trip if no xbars */
  if (fp->num_xbars == 0) return 0;

  /* Attach some scratch memory to each xbar for us to use */
  rc = lf_clos_alloc_space(fp);
  if (rc == -1) LF_ERROR(("Error allocating scratch xbar space"));
  alloced = TRUE;

  /* zero out everyone's max distance */
  FOR_EACH_XBAR(fp, xp)
    xp->min_nic_dist = -1;
    LC_XBAR(xp)->max_nic_dist = 0;
    xp->clos_level = 0;
    LC_XBAR(xp)->last_visitor = NULL;
  END_XBAR_LOOP

  /* find closest and farthest NIC to each xbar */
  FOR_EACH_XBAR(fp, xp)
    lf_find_min_nic_dist(xp);
  END_XBAR_LOOP

  /* clear last visitors, and now find the max distance to any NIC */
  FOR_EACH_XBAR(fp, xp)
    LC_XBAR(xp)->last_visitor = NULL;
  END_XBAR_LOOP
  FOR_EACH_XBAR(fp, xp)
    lf_find_max_nic_dist(xp);
  END_XBAR_LOOP

  /* Find the clos depth */
  LC_FABRIC(fp)->clos_depth = lf_find_clos_depth(fp);

  /* Propagate clos_depth to all xbars */
  rc = lf_propagate_clos_levels(fp);
  if (rc == -1) LF_ERROR(("Error propagsting clos levels"));

  if (0) {
    LF_CALLOC(links, int, LC_FABRIC(fp)->clos_depth+1);

    FOR_EACH_XBAR(fp, xp)
      int l;
      int p;

      l = xp->clos_level;

      /* validate l, as this is just informational */
      if (l > LC_FABRIC(fp)->clos_depth || l < 0) {
	LF_FREE(links);
	LF_ERROR(("Bad clos-level on xbar, not a clos?"));
	continue;
      }

      /* check every xbar neighbor */
      for (p=0; p<xp->num_ports; ++p) {
	struct lf_xbar *xp2;

	xp2 = LF_XBAR(xp->topo_ports[p]);
	if (xp2 != NULL
	    && (xp2->ln_type == LF_NODE_NIC
	        || (xp2->ln_type == LF_NODE_XBAR
		    && xp2->clos_level < l))) {
	  ++links[l];
	}
      }
    END_XBAR_LOOP
  }

 except:
  if (alloced) lf_clos_free_space(fp);
  LF_FREE(links);
  return rc;
}

/*
 * Find the distance to the closest NIC from this xbar
 */
static void
lf_find_min_nic_dist(
  struct lf_xbar *xbar)
{
  int dist;
  struct lf_xbar *xp;
  struct lf_xbar *work_list;

  dist = 1;

  /* initialize work_list to just be first xbar */
  work_list = xbar;
  LC_XBAR(xbar)->bf_next = NULL;
  LC_XBAR(xbar)->last_visitor = xbar;

  /* Breadth-first until nothing more to do */
  while (work_list != NULL) {

    /* save this list and start building new work_list */
    xp = work_list;
    work_list = NULL;

    /* process every xbar in old work_list */
    while (xp != NULL) {
      int p;

      /* Check all ports for NICs and unvisited xbars */
      for (p=0; p<xp->num_ports; ++p) {
	union lf_node *np2;

	np2 = xp->topo_ports[p];
	if (np2 == NULL) continue;

	if (np2->ln_type == LF_NODE_NIC) {
	  xbar->min_nic_dist = dist;
	  return;

	} else if (np2->ln_type == LF_NODE_XBAR
		   && LC_XBAR(np2)->last_visitor != xbar) {

	  /* add to work_list */
	  LC_XBAR(np2)->bf_next = work_list;
	  work_list = LF_XBAR(np2);
	  
	  /* mark as visited to prevent multiple visitors */
	  LC_XBAR(np2)->last_visitor = xbar;
	}
      }

      /* next in list */
      xp = LC_XBAR(xp)->bf_next;
    }

    /* increment distance for each go-round */
    ++dist;
  }
}

/*
 * Find the longest minimum distance from this xbar to any NIC by using bfs
 * until there are no more xbars.  
 */
static void
lf_find_max_nic_dist(
  struct lf_xbar *xbar)
{
  int dist;
  struct lf_xbar *xp;
  struct lf_xbar *work_list;

  dist = 1;

  /* initialize work_list to just be first xbar */
  work_list = xbar;
  LC_XBAR(xbar)->bf_next = NULL;
  LC_XBAR(xbar)->last_visitor = xbar;

  /* Breadth-first until nothing more to do */
  while (work_list != NULL) {

    /* save this list and start building new work_list */
    xp = work_list;
    work_list = NULL;

    /* process every xbar in old work_list */
    while (xp != NULL) {
      int p;

      /* Check all ports for NICs and unvisited xbars */
      for (p=0; p<xp->num_ports; ++p) {
	union lf_node *np2;

	np2 = xp->topo_ports[p];
	if (np2 == NULL) continue;

	if (np2->ln_type == LF_NODE_NIC) {
	  if (dist > LC_XBAR(xbar)->max_nic_dist) {
	    LC_XBAR(xbar)->max_nic_dist = dist;
	  }

	} else if (np2->ln_type == LF_NODE_XBAR
		   && LC_XBAR(np2)->last_visitor != xbar) {
	  struct lf_xbar *xp2;

	  xp2 = LF_XBAR(np2);

	  /* add to new work_list */
	  LC_XBAR(xp2)->bf_next = work_list;
	  work_list = xp2;
	  
	  /* mark as visited to prevent multiple visitors */
	  LC_XBAR(xp2)->last_visitor = xbar;
	}
      }

      /* next in list */
      xp = LC_XBAR(xp)->bf_next;
    }

    /* increment distance for each go-round */
    ++dist;
  }
}

/*
 * propagate clos levels.  start with all xbars whose "max_nic_dist" is
 * equal to the fabric clos depth.  Set their clos_level to be clos_depth.
 * Then, propagate out to all other xbars.  
 * We may go down then back up to tag "false spines", so we need to be
 * willing to go both "up" and "down" - if "max_distance" is greater than
 * ours, subtract one from our clos level, if greater, add one.
 * If equal, print an error because I am not sure that's possible...
 */
static int
lf_propagate_clos_levels(
  struct lf_fabric *fp)
{
  struct lf_xbar *work_list;
  int x;

  /* initialize work_list to be all xbars with max_nic_dist == clos_depth */
  work_list = NULL;
  for (x=0; x<fp->num_xbars; ++x) {
    struct lf_xbar *xp;

    xp = fp->xbars[x];
    if (LC_XBAR(xp)->max_nic_dist == LC_FABRIC(fp)->clos_depth) {
      xp->clos_level = LC_FABRIC(fp)->clos_depth;

      /* tag this as visited and add to work list */
      LC_XBAR(xp)->last_visitor = fp;
      LC_XBAR(xp)->bf_next = work_list;
      work_list = xp;
    }
  }

  /* Breadth-first until nothing more to do */
  while (work_list != NULL) {
    struct lf_xbar *xp;

    /* save this list and start building new work_list */
    xp = work_list;
    work_list = NULL;

      /* process every xbar in old work_list */
    while (xp != NULL) {
      int p;

      /* Check all ports for unvisited xbars */
      for (p=0; p<xp->num_ports; ++p) {
	struct lf_xbar *xp2;

	xp2 = LF_XBAR(xp->topo_ports[p]);

	/* if non-existant, non-xbar, or visited, skip it */
	if (xp2 == NULL
	    || xp2->ln_type != LF_NODE_XBAR
	    || LC_XBAR(xp2)->last_visitor == fp) {
	  continue;
	}

	/* set neighbors' clos_level based on mine */
	if (LC_XBAR(xp2)->max_nic_dist < LC_XBAR(xp)->max_nic_dist) {
	  xp2->clos_level = xp->clos_level + 1;
	} else {
	  xp2->clos_level = xp->clos_level - 1;
	}

	/* add to new work_list */
	LC_XBAR(xp2)->bf_next = work_list;
	work_list = xp2;
	
	/* mark as visited to prevent multiple visitors */
	LC_XBAR(xp2)->last_visitor = fp;
      }

      /* next in list */
      xp = LC_XBAR(xp)->bf_next;
    }
  }

  /* We could cleanup a little here because in a "flat-top" clos,
   * there will end up being no "level 1" nodes.  We could find the 
   * "minimum" clos level anywhere and subtract from everything
   * to make the minimum 1, but it's not necessary yet.
   */
  return 0;
}

/*
 * Find the clos depth by starting from some host, and traveling to xbars
 * with decreasing "max distance", choosing each time the largest max distance
 * which is less than our own until we can go no farther.  The max_nic_distance
 * of that xbar is the depth of the clos.  This allows us to avoid finding 
 * "false spines" which are sub-spines within the overall fabric.
 */
static int
lf_find_clos_depth(
  struct lf_fabric *fp)
{
  struct lf_nic *nicp;
  struct lf_xbar *xp;

  /* just take the first connected NIC in the fabric.
   * If really a clos, any will do.
   */
  {
    int h;

    nicp = NULL;
    xp = NULL;
    for (h=0; h<fp->num_hosts; ++h) {
      if (fp->hosts[h]->num_nics > 0) {
	nicp = fp->hosts[h]->nics[0];
	xp = LF_XBAR(nicp->topo_ports[0]);
	if (xp != NULL && xp->ln_type == LF_NODE_XBAR) {
	  break;
	}
      }
    }
    if (xp == NULL) return 0;
  }

  while (1) {
    struct lf_xbar *nxp;
    int max_max;
    int p;

    max_max = -1;
    for (p=0; p<xp->num_ports; ++p) {
      union lf_node *onp;

      onp = xp->topo_ports[p];
      if (onp != NULL
	  && onp->ln_type == LF_NODE_XBAR
	  && LC_XBAR(onp)->max_nic_dist < LC_XBAR(xp)->max_nic_dist
	  && LC_XBAR(onp)->max_nic_dist > max_max) {
	max_max = LC_XBAR(onp)->max_nic_dist;
	nxp = LF_XBAR(onp);
      }
    }

    /*
     * If no higher nodes found, then this is a spine xbar!
     */
    if (max_max == -1) {
      return LC_XBAR(xp)->max_nic_dist;

    } else {
      xp = nxp;		/* otherwise, move on to the new xbar */
    }
  }
}

/*
 * Alloacate space for working variables
 */
static int
lf_clos_alloc_space(
  struct lf_fabric *fp)
{
  struct lf_clos_level_fabric_data *fdp;
  struct lf_clos_level_xbar_data *xdp;
  struct lf_xbar *xp;

  fdp = NULL;
  xdp = NULL;

  /* allocate data space (all xbars are in one alloced array for efficiency) */
  LF_CALLOC(fdp, struct lf_clos_level_fabric_data, 1);
  LF_CALLOC(xdp, struct lf_clos_level_xbar_data, fp->num_xbars);

  /* Save old fabric user data ptr and add ours in */
  fdp->old_user = fp->user.v;
  fp->user.v = fdp;

  FOR_EACH_XBAR(fp,xp)
    xdp[__x].old_user = xp->user.v;
    xp->user.v = xdp + __x;
  END_XBAR_LOOP

  return 0;

 except:
  LF_FREE(fdp);
  LF_FREE(xdp);
  return -1;
}

/*
 * Free out temporary working space
 */
static void
lf_clos_free_space(
  struct lf_fabric *fp)
{
  struct lf_clos_level_fabric_data *fdp;
  struct lf_clos_level_xbar_data *xdp;
  struct lf_xbar *xp;

  /* Restore old fabric user data ptr and free ours */
  fdp = fp->user.v;
  fp->user.v = fdp->old_user;
  LF_FREE(fdp);

  /* first xbar has the address we will free */
  xdp = fp->xbars[0]->user.v;
  FOR_EACH_XBAR(fp,xp)
    xp->user.v = xdp[__x].old_user;
  END_XBAR_LOOP
  LF_FREE(xdp);
}
